/*
 * Connect3D - Reconstruction of water-tight triangulated meshes from unstructured point clouds
 *
 * please credit the following article: Stefan Ohrhallinger, Sudhir Mudur and Michael Wimmer:
 * 'Minimizing Edge Length to Connect Sparsely Sampled Unstructured Point Sets',
 * Shape Modeling International 2013, published in Computers & Graphics Journal, 2013.
 *
 * Copyright (C) 2013 Stefan Ohrhallinger, Daniel Prieler
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 */
 
#include <boost/algorithm/string.hpp>

#include "precompiled.h"

#include "c3dDataStructures.h"
#include "ply.h"

#include "FileIO.h"


typedef struct PlyFace {
	unsigned char intensity; /* this user attaches intensity to faces */
	unsigned char nverts; /* number of vertex indices in list */
	int *verts; /* vertex index list */
} PlyFace;

typedef struct PlyVertex {
	float x, y, z; /* the usual 3-space position of a vertex */
} PlyVertex;

PlyProperty face_props[] = { /* list of property information for a vertex */
{ "intensity", PLY_UCHAR, PLY_UCHAR, offsetof(PlyFace,intensity), 0, 0, 0, 0 }, { "vertex_index", PLY_INT, PLY_INT,
		offsetof(PlyFace,verts), 1, PLY_UCHAR, PLY_UCHAR, offsetof(PlyFace,nverts) },
};

PlyProperty vert_props[] = { /* list of property information for a vertex */
{ "x", PLY_FLOAT, PLY_FLOAT, offsetof(PlyVertex,x), 0, 0, 0, 0 }, { "y", PLY_FLOAT, PLY_FLOAT, offsetof(PlyVertex,y), 0, 0, 0, 0 }, { "z", PLY_FLOAT, PLY_FLOAT, offsetof(PlyVertex,z), 0, 0, 0, 0 },
};




 VertexData *GeometryLoader::loadFile(const std::string &filename)
{
	//check for extension
	size_t fLength = filename.length();
	std::string ext = filename.substr(fLength - 4, 4);
	std::string filePath = filename;


	std::cout << "Loading filepath " << filePath << std::endl;

	if(boost::iequals(ext, ".ply"))
	{
		return readPlyFile(filePath);
	}
	else if(boost::iequals(ext, ".pts"))
	{
		return readPtsFile(filePath);
	}
	else
	{
		std::cout << "Couldn't handle extension of file " << filename << std::endl;
		return nullptr;
	}

}



 VertexData *GeometryLoader::readPlyFile(const std::string &path)
{
	int i, k;
	PlyFile *ply;
	int nelems;
	char **elist;
	int file_type;
	float version;
	int nprops;
 	int num_elems;
	bool error = false;

	char *elem_name;

	char *filePath = const_cast<char *>(path.c_str());


	/* open a PLY file for reading */
	ply = ply_open_for_reading(filePath, &nelems, &elist, &file_type, &version);

	if (!ply) {
		std::cout << "error opening PLY file" << std::endl;
		error = true;
        return nullptr;
	}

	 VertexData *gData = new  VertexData();
	
	for (k = 0; k < nelems; k++) {

		/* get the description of the first element */
		elem_name = elist[k];
		//char *withoutconst = (char*)elem_name;
        	ply_get_element_description(ply, elem_name, &num_elems, &nprops);

		/* if we're on vertex elements, read them in */
		if (equal_strings("vertex", elem_name)) {

 			gData->numVertices = static_cast<size_t>(num_elems);

			gData->vertices.reset(new Vector3D[num_elems]);

			/* set up for getting vertex elements */

			ply_get_property(ply, elem_name, &vert_props[0]);
			ply_get_property(ply, elem_name, &vert_props[1]);
			ply_get_property(ply, elem_name, &vert_props[2]);

			for (i = 0; static_cast<size_t>(i) < gData->numVertices; i++) {
				ply_get_element(ply, (void *)&gData->vertices[i]);
			}
		}

	}

	/* close the PLY file */
	ply_close(ply);

	if(error)
	{
		delete gData;
		return nullptr;
	}
	
	return gData;
}

 VertexData *GeometryLoader::readPtsFile(const std::string &path)
{
	size_t i;
	std::vector<Vector3D> points;

	std::ifstream file(path);

	if(!file.is_open())
	{
		std::cout << "Couldn't open file " << path << std::endl;
		return nullptr;
	}

	 VertexData *gData = new  VertexData;

	while (!file.eof())
	{
		Vector3D vec;
		file >> vec[0] >> vec[1] >> vec[2];
		points.push_back(vec);
	}

	gData->numVertices = points.size() - 1;
	gData->vertices.reset(new Vector3D[gData->numVertices]);
	
	Vector3D vertex;

	for (i = 0; i < gData->numVertices; i++)
	{
		vertex = points.at(i);
		gData->vertices[i] = vertex;
	}

	points.clear();

	return gData;
	
}


 bool GeometryWriter::testWrite(const std::string &fileName)
 {
	 std::ofstream fileOut(fileName);
	 return fileOut.is_open();
 }


/*
 * write triangulation to OFF-file
 */
bool GeometryWriter::writeOFFFile(const std::string &fileName, const std::vector<Triangle *> &triangles, const Dt *dTriangulation, const Vector3D *vertexData)
{
	int i;

	std::string last4 = fileName.substr(fileName.size() - 4, 4);
	
	std::string filePath;
	if(last4 == ".off")
		filePath = fileName;
	else
		filePath = fileName + ".off";

	std::ofstream fileOut(filePath);

	if (fileOut.is_open())
	{
		// count triangles
		int triangleCount = 0;

		for (std::vector<Triangle *>::const_iterator iter = triangles.begin(); iter != triangles.end(); iter++)
		{
			if ((*iter)->exists())
				triangleCount++;
		}

		// write OFF header
		fileOut << "OFF" << std::endl;
		fileOut << (int)dTriangulation->number_of_vertices() << " " << triangleCount << " 0" << std::endl;

		// write vertex list
		for (Dt::Finite_vertices_iterator vhIter = dTriangulation->finite_vertices_begin(); vhIter != dTriangulation->finite_vertices_end(); vhIter++)
		{
			DVertex currVH = vhIter;
			int vIndex = currVH->info().index;

			if (vIndex != -1)
			{
				Vector3D vec = vertexData[vIndex];

				for (i = 0; i < 3; i++)
					fileOut << vec[i] << " ";

				fileOut << std::endl;

				if (vIndex >= (int)dTriangulation->number_of_vertices())
				{
					std::cout << "v #" << vIndex << ">= " << (int)dTriangulation->number_of_vertices() << "(vertexcount)!" << std::endl;
					assert(false);
				}
			}
		}

		// write triangle list
		for (std::vector<Triangle *>::const_iterator iter = triangles.begin(); iter != triangles.end(); iter++)
		{
			if ((*iter)->exists())
			{
				fileOut << "3 ";

				for (i = 0; i < 3; i++)
					fileOut << (*iter)->vertex(i)->info().index << " ";

				fileOut << std::endl;
			}
		}

		fileOut.close();
	}
	else
	{
		std::cerr << "ERROR: could not write OFF file " << filePath << std::endl;
		return false;
	}

	return true;
}


