/*
 * Connect3D - Reconstruction of water-tight triangulated meshes from unstructured point clouds
 *
 * please credit the following article: Stefan Ohrhallinger, Sudhir Mudur and Michael Wimmer:
 * 'Minimizing Edge Length to Connect Sparsely Sampled Unstructured Point Sets',
 * Shape Modeling International 2013, published in Computers & Graphics Journal, 2013.
 *
 * Copyright (C) 2013 Stefan Ohrhallinger, Daniel Prieler
 * This program is free software; you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 */
 
#include "precompiled.h"
#include "c3dDataStructures.h"
#include "FileIO.h"
#include "connect3d.h"


const static std::string Default_Filename = "Reconstructed.off";
void printUsage()
{
	std::cout << "USAGE:\nconnect3d infile [outfile]\n\nThe input filename has to end either in:\n";
	std::cout << "- '.ply' (and is then interpreted as a stanford polygon PLY file, only point coordinates are read in) or in\n";
	std::cout << "- '.pts', which contains only the point coordinates as ASCII triples of floats per line as follows:\n";
	std::cout << "  x0 y0 z0\n";
	std::cout << "  x1 y1 z1\n";
	std::cout << "  ...\n\n";
	std::cout << "If no outfile is given, the reconstructed model is saved to Reconstructed.off in the current directory." << std::endl;
}

bool progressCallback(const int pos, const std::string &msg)
{
	std::cout << pos << " % completed.. Current step: " << msg << std::endl;
	return true;
}

int main(int argc, char **argv)
{
	if(argc < 2)
	{
		printUsage();
		return EXIT_FAILURE;
	}

	std::cout << Connect3D::info() << "\n" << std::endl;

	std::string pathIn(argv[1]);

	std::string pathOut = Default_Filename;

	
	if(argc > 2)
		pathOut = std::string(argv[2]);

	std::cout << "Infile: " << pathIn << " and outfile: " << pathOut << std::endl;

	VertexData *geometry = GeometryLoader::loadFile(pathIn);

	if(geometry == nullptr)
	{
		std::cout << "Exiting" << std::endl;
		printUsage();
		return EXIT_FAILURE;
	}

	if(GeometryWriter::testWrite(pathOut) == false)
	{
		std::cerr << "Can't write to file: " << pathOut << ". Maybe need different permissions." << std::endl;
		delete geometry;
		return EXIT_FAILURE;
	}

	c3dModelData data;
	data.numVertices = geometry->numVertices;
	data.vertices.reset(geometry->vertices.get());

	std::cout << "Loading complete, starting reconstruction.." << std::endl;

	std::vector<Triangle *> tris = Connect3D::reconstruct(data, &progressCallback);
	
	bool writeSuccess = GeometryWriter::writeOFFFile(pathOut, tris, data.dt.get(), data.vertices.get());
	if(writeSuccess != true)
	{
		std::cerr << "Error writing mesh to file: " << pathOut << std::endl;
		return EXIT_FAILURE;
	}

	std::cout << "Mesh reconstruction complete, wrote model to: " << pathOut << "\n" << std::endl;
	return EXIT_SUCCESS;

}
